@@ -8,6 +8,7 @@
 #
 # The above copyright notice and this permission notice shall be included in all copies or
 # substantial portions of the Software.
+import datetime
 import gzip
 import json
 import mmap
@@ -32,18 +33,7 @@ from pathlib import Path
 from subprocess import PIPE, Popen
 from tempfile import TemporaryFile
 from textwrap import dedent, indent
-from typing import (
-    IO,
-    TYPE_CHECKING,
-    Any,
-    Dict,
-    Iterable,
-    List,
-    Optional,
-    Tuple,
-    Type,
-    Union,
-)
+from typing import IO, TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, Type, Union
 
 from google.api_core.exceptions import Conflict
 from google.cloud import bigquery, bigquery_storage_v1, storage
@@ -79,11 +69,17 @@ PARTITION_STRATEGY = {
 @dataclass
 class BigQueryTable:
     name: str
+    """The name of the table."""
     dataset: str
+    """The dataset that this table belongs to."""
     project: str
+    """The project that this table belongs to."""
     jsonschema: Dict[str, Any]
+    """The jsonschema for this table."""
     ingestion_strategy: IngestionStrategy
+    """The ingestion strategy for this table."""
     transforms: Dict[str, bool] = field(default_factory=dict)
+    """A dict of transformation rules to apply to the table schema."""
 
     @property
     def schema_translator(self) -> "SchemaTranslator":
@@ -101,13 +97,11 @@ class BigQueryTable:
             return self.schema_translator.translated_schema_transformed
         return self.schema_translator.translated_schema
 
-    def get_resolved_schema(
-        self, apply_transforms: bool = False
-    ) -> List[bigquery.SchemaField]:
+    def get_resolved_schema(self, apply_transforms: bool = False) -> List[bigquery.SchemaField]:
         """Returns the schema for this table after factoring in the ingestion strategy."""
-        if self.ingestion_strategy == IngestionStrategy.FIXED:
+        if self.ingestion_strategy is IngestionStrategy.FIXED:
             return DEFAULT_SCHEMA
-        elif self.ingestion_strategy == IngestionStrategy.DENORMALIZED:
+        elif self.ingestion_strategy is IngestionStrategy.DENORMALIZED:
             return self.get_schema(apply_transforms)
 
     def __str__(self) -> str:
@@ -143,15 +137,18 @@ class BigQueryTable:
         apply_transforms: bool = False,
         **kwargs,
     ) -> Tuple[bigquery.Dataset, bigquery.Table]:
-        """Creates a dataset and table for this table."""
+        """Creates a dataset and table for this table.
+
+        This is a convenience method that wraps the creation of a dataset and
+        table in a single method call. It is idempotent and will not create
+        a new table if one already exists."""
         if not hasattr(self, "_dataset"):
             self._dataset = client.create_dataset(self.as_dataset(), exists_ok=True)
         if not hasattr(self, "_table"):
             try:
                 self._table = client.create_table(
                     self.as_table(
-                        apply_transforms
-                        and self.ingestion_strategy != IngestionStrategy.FIXED,
+                        apply_transforms and self.ingestion_strategy is not IngestionStrategy.FIXED,
                         **kwargs,
                     )
                 )
@@ -167,11 +164,14 @@ class BigQueryTable:
         schema_dump = json.dumps(self.jsonschema)
         return {
             "clustering_fields": ["_sdc_batched_at"],
-            "description": "Generated by target-bigquery.\nStream Schema\n{schema}\nBigQuery Ingestion Strategy: {strategy}".format(
-                schema=(schema_dump[:16000] + "...")
-                if len(schema_dump) > 16000
-                else schema_dump,
-                strategy=self.ingestion_strategy,
+            "description": (
+                "Generated by target-bigquery.\nStream Schema\n{schema}\nBigQuery Ingestion"
+                " Strategy: {strategy}".format(
+                    schema=(schema_dump[:16000] + "...")
+                    if len(schema_dump) > 16000
+                    else schema_dump,
+                    strategy=self.ingestion_strategy,
+                )
             ),
             "time_partitioning": TimePartitioning(
                 type_=TimePartitioningType.MONTH, field="_sdc_batched_at"
@@ -179,9 +179,7 @@ class BigQueryTable:
         }
 
     def __hash__(self) -> int:
-        return hash(
-            (self.name, self.dataset, self.project, json.dumps(self.jsonschema))
-        )
+        return hash((self.name, self.dataset, self.project, json.dumps(self.jsonschema)))
 
 
 @dataclass
@@ -232,6 +230,7 @@ class BaseBigQuerySink(BatchSink):
         schema: Dict[str, Any],
         key_properties: Optional[List[str]],
     ) -> None:
+        """Initialize the sink."""
         super().__init__(target, stream_name, schema, key_properties)
         self._credentials = BigQueryCredentials(
             self.config.get("credentials_path"),
@@ -239,17 +238,32 @@ class BaseBigQuerySink(BatchSink):
             self.config["project"],
         )
         self.client = bigquery_client_factory(self._credentials)
-        self.table = BigQueryTable(
-            name=self.table_name,
-            dataset=self.config["dataset"],
-            project=self.config["project"],
-            jsonschema=self.schema,
-            transforms=self.config.get("column_name_transforms", {}),
-            ingestion_strategy=self.ingestion_strategy,
-        )
-
-        self.create_target(key_properties)
+        opts = {
+            "project": self.config["project"],
+            "dataset": self.config["dataset"],
+            "jsonschema": self.schema,
+            "transforms": self.config.get("column_name_transforms", {}),
+            "ingestion_strategy": self.ingestion_strategy,
+        }
+        self.table = BigQueryTable(name=self.table_name, **opts)
+        self.create_target(key_properties=key_properties)
         self.update_schema()
+        self.merge_target: Optional[BigQueryTable] = None
+        if (
+            key_properties
+            and self.ingestion_strategy is IngestionStrategy.DENORMALIZED
+            and self.config.get("dedupe")
+        ):
+            from copy import copy
+
+            self.merge_target = copy(self.table)
+            self.table = BigQueryTable(name=f"{self.table_name}__{int(time.time())}", **opts)
+            self.table.create_table(
+                self.client,
+                self.apply_transforms,
+                expires=datetime.datetime.now() + datetime.timedelta(days=1),
+            )
+            time.sleep(2.5)  # Wait for eventual consistency
         self.global_par_typ = target.par_typ
         self.global_queue = target.queue
         self.increment_jobs_enqueued = target.increment_jobs_enqueued
@@ -279,9 +293,7 @@ class BaseBigQuerySink(BatchSink):
     def _validate_and_parse(self, record: dict) -> dict:
         return record
 
-    def preprocess_record(
-        self, record: Dict[str, Any], context: Dict[str, Any]
-    ) -> Dict[str, Any]:
+    def preprocess_record(self, record: Dict[str, Any], context: Dict[str, Any]) -> Dict[str, Any]:
         """Preprocess a record before writing it to the sink."""
         metadata = {
             k: record.pop(k, None)
@@ -337,6 +349,43 @@ class BaseBigQuerySink(BatchSink):
         """Return a worker class for the given parallelization type."""
         raise NotImplementedError
 
+    def clean_up(self) -> None:
+        """Clean up the target table."""
+        if self.merge_target is not None:
+            # We must merge the temp table into the target table.
+            target = self.merge_target.as_table()
+            date_columns = ["_sdc_extracted_at", "_sdc_received_at"]
+            tmp, ctas_tmp = None, "SELECT 1 AS _no_op"
+            if self.config.get("non_unique_key_properties", True):
+                # We can't use MERGE with a non-unique key, so we need to dedupe the temp table into
+                # a _SESSION scoped intermediate table.
+                tmp = f"{target.name}__tmp"
+                dedupe_query = (
+                    f"SELECT * FROM {self.table} "
+                    f"QUALIFY ROW_NUMBER() OVER (PARTITION BY {', '.join(self.key_properties)} "
+                    f"ORDER BY COALESCE({', '.join(date_columns)}) DESC) = 1"
+                )
+                ctas_tmp = f"CREATE OR REPLACE TEMP TABLE {tmp} AS {dedupe_query}"
+            merge_clause = (
+                f"MERGE `{self.merge_target}` AS target USING `{tmp or self.table}` AS source ON "
+                + " AND ".join(f"target.`{f}` = source.`{f}`" for f in self.key_properties)
+            )
+            update_clause = "UPDATE SET " + ", ".join(
+                f"target.`{f.name}` = source.`{f.name}`" for f in target.schema
+            )
+            insert_clause = (
+                f"INSERT ({', '.join(f.name for f in target.schema)}) "
+                f"VALUES ({', '.join(f'source.`{f.name}`' for f in target.schema)})"
+            )
+            self.client.query(
+                f"{ctas_tmp}; {merge_clause} "
+                f"WHEN MATCHED THEN {update_clause} "
+                f"WHEN NOT MATCHED THEN {insert_clause}; "
+                f"DROP TABLE IF EXISTS {self.table};"
+            ).result()
+            self.table = self.merge_target
+            self.merge_target = None
+
 
 class Denormalized:
     """This class provides common overrides for denormalized sinks and should be subclassed
@@ -411,9 +460,7 @@ def storage_client_factory(
 ) -> bigquery_storage_v1.BigQueryWriteClient:
     """Get a BigQuery Storage Write client."""
     if creds.path:
-        return bigquery_storage_v1.BigQueryWriteClient.from_service_account_file(
-            creds.path
-        )
+        return bigquery_storage_v1.BigQueryWriteClient.from_service_account_file(creds.path)
     elif creds.json:
         return bigquery_storage_v1.BigQueryWriteClient.from_service_account_info(
             json.loads(creds.json)
@@ -480,15 +527,15 @@ class SchemaTranslator:
     def generate_view_statement(self, table_name: str) -> str:
         """Generate a CREATE VIEW statement for the SchemaTranslator `schema`."""
         projection = ""
-        for field in self.translated_schema[:]:
-            if field.mode == "REPEATED":
-                projection += indent(
-                    self._wrap_json_array(field, path="$", depth=1), " " * 4
-                )
+        for field_ in self.translated_schema[:]:
+            if field_.mode == "REPEATED":
+                projection += indent(self._wrap_json_array(field_, path="$", depth=1), " " * 4)
             else:
-                projection += indent(self._bigquery_field_to_projection(field), " " * 4)
+                projection += indent(self._bigquery_field_to_projection(field_), " " * 4)
 
-        return f"CREATE OR REPLACE VIEW {table_name}_view AS \nSELECT \n{projection} FROM {table_name}"
+        return (
+            f"CREATE OR REPLACE VIEW {table_name}_view AS \nSELECT \n{projection} FROM {table_name}"
+        )
 
     def _jsonschema_property_to_bigquery_column(
         self, name: str, schema_property: dict
@@ -505,14 +552,12 @@ class SchemaTranslator:
             property_format = schema_property.get("format", None)
 
         if "array" in property_type:
+            if "items" not in schema_property:
+                return SchemaField(name, "JSON", "REPEATED")
             items_schema: dict = schema_property["items"]
-            items_type = bigquery_type(
-                items_schema["type"], items_schema.get("format", None)
-            )
+            items_type = bigquery_type(items_schema["type"], items_schema.get("format", None))
             if items_type == "record":
-                return self._translate_record_to_bigquery_schema(
-                    name, items_schema, "REPEATED"
-                )
+                return self._translate_record_to_bigquery_schema(name, items_schema, "REPEATED")
             return SchemaField(name, items_type, "REPEATED")
         elif "object" in property_type:
             return self._translate_record_to_bigquery_schema(name, schema_property)
@@ -531,18 +576,15 @@ class SchemaTranslator:
         return SchemaField(name, "RECORD", mode, fields=fields)
 
     def _bigquery_field_to_projection(
-        self,
-        field: SchemaField,
-        path: str = "$",
-        depth: int = 0,
-        base: str = "data",
-        rebase: bool = False,
+        self, field: SchemaField, path: str = "$", depth: int = 0, base: str = "data"
     ) -> str:
         """Translate a BigQuery schema field into a SQL projection."""
         # Pass-through _sdc columns into the projection as-is
         if field.name.startswith("_sdc_"):
             return f"{field.name},\n"
-        scalar = f"{base}.{field.name}" if not rebase else f"{base}"
+
+        scalar = f"{base}.{field.name}"
+        from_base = f"{path}.{field.name}" if base == "data" else "$"
 
         # Records are handled recursively
         if field.field_type.upper() == "RECORD":
@@ -552,12 +594,10 @@ class SchemaTranslator:
                     "".join(
                         [
                             self._bigquery_field_to_projection(
-                                f, f"{path}.{field.name}", depth + 1
+                                f, path=from_base, depth=depth + 1, base=base
                             )
                             if not f.mode == "REPEATED"
-                            else self._wrap_json_array(
-                                f, f"{path}.{field.name}", depth, base
-                            )
+                            else self._wrap_json_array(f, path=from_base, depth=depth, base=base)
                             for f in field.fields
                         ]
                     ).rstrip(",\n"),
@@ -567,20 +607,16 @@ class SchemaTranslator:
             )
         # Nullable fields require a JSON_VALUE call which creates a 2-stage cast
         elif field.is_nullable:
-            return (" " * depth * 2) + self._wrap_nullable_json_value(
-                field, f"{path}.{field.name}", base
-            )
+            return (" " * depth * 2) + self._wrap_nullable_json_value(field, path=path, base=base)
         # These are not nullable so if the type is known, we can do a 1-stage extract & cast
         elif field.field_type.upper() == "STRING":
             return (
-                (" " * depth * 2)
-                + f"STRING({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
-            )
+                " " * depth * 2
+            ) + f"STRING({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
         elif field.field_type.upper() == "INTEGER":
             return (
-                (" " * depth * 2)
-                + f"INT64({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
-            )
+                " " * depth * 2
+            ) + f"INT64({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
         elif field.field_type.upper() == "FLOAT":
             return (
                 (" " * depth * 2)
@@ -588,21 +624,18 @@ class SchemaTranslator:
             )
         elif field.field_type.upper() == "BOOLEAN":
             return (
-                (" " * depth * 2)
-                + f"BOOL({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
-            )
+                " " * depth * 2
+            ) + f"BOOL({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
         # Fallback to a 2-stage extract & cast
         else:
-            return (" " * depth * 2) + self._wrap_nullable_json_value(
-                field, f"{path}.{field.name}" if not rebase else "$", base
-            )
+            return (" " * depth * 2) + self._wrap_nullable_json_value(field, path, base)
 
     def _wrap_json_array(
         self, field: SchemaField, path: str, depth: int = 0, base: str = "data"
     ) -> str:
         """Translate a BigQuery schema field into a SQL projection for a repeated field."""
         v = self._bigquery_field_to_projection(
-            field, "$", depth, f"{field.name}__rows", rebase=True
+            field, path="$", depth=depth, base=f"{field.name}__rows"
         ).rstrip(", \n")
         return (" " * depth * 2) + indent(
             dedent(
@@ -625,18 +658,25 @@ class SchemaTranslator:
         """Translate a BigQuery schema field into a SQL projection for a nullable field."""
         typ = field.field_type.upper()
         if typ == "STRING":
-            return f"JSON_VALUE({base}, '{path}') as {transform_column_name(field.name, **self.transforms)},\n"
+            return (
+                f"JSON_VALUE({base}, '{path}.{field.name}') as"
+                f" {transform_column_name(field.name, **self.transforms)},\n"
+            )
         if typ == "FLOAT":
             typ = "FLOAT64"
         if typ in ("INT", "INTEGER"):
             typ = "INT64"
-        return f"CAST(JSON_VALUE({base}, '{path}') as {typ}) as {transform_column_name(field.name, **self.transforms)},\n"
+        return (
+            f"CAST(JSON_VALUE({base}, '{path}.{field.name}') as {typ}) as"
+            f" {transform_column_name(field.name, **self.transforms)},\n"
+        )
 
 
 class Compressor:
     """Compresses streams of bytes using gzip."""
 
     def __init__(self) -> None:
+        """Initialize the compressor."""
         self._compressor = None
         self._closed = False
         if shutil.which("gzip") is not None:
@@ -650,15 +690,18 @@ class Compressor:
             self._gzip = gzip.GzipFile(fileobj=self._buffer, mode="wb")
 
     def write(self, data: bytes) -> None:
+        """Write data to the compressor."""
         if self._closed:
             raise ValueError("I/O operation on closed compressor.")
         self._gzip.write(data)
 
     def flush(self) -> None:
+        """Flush the compressor buffer."""
         self._gzip.flush()
         self._buffer.flush()
 
     def close(self) -> None:
+        """Close the compressor and wait for the gzip process to finish."""
         if self._closed:
             return
         self._gzip.close()
@@ -669,6 +712,7 @@ class Compressor:
         self._closed = True
 
     def getvalue(self) -> bytes:
+        """Return the compressed buffer as a bytes object."""
         if not self._closed:
             self.close()
         if self._compressor is not None:
@@ -676,6 +720,7 @@ class Compressor:
         return self._buffer.getvalue()
 
     def getbuffer(self) -> Union[memoryview, mmap.mmap]:
+        """Return the compressed buffer as a memoryview or mmap."""
         if not self._closed:
             self.close()
         if self._compressor is not None:
@@ -684,9 +729,11 @@ class Compressor:
 
     @property
     def buffer(self) -> IO[bytes]:
+        """Return the compressed buffer as a file-like object."""
         return self._buffer
 
     def __del__(self) -> None:
+        """Close the compressor and wait for the gzip process to finish. Dereference the buffer."""
         if not self._closed:
             self.close()
         # close the buffer, ignore error if we have an incremented rc due to memoryview
@@ -703,9 +750,7 @@ class Compressor:
 
 
 # pylint: disable=no-else-return,too-many-branches,too-many-return-statements
-def bigquery_type(
-    property_type: List[str], property_format: Optional[str] = None
-) -> str:
+def bigquery_type(property_type: List[str], property_format: Optional[str] = None) -> str:
     """Convert a JSON Schema type to a BigQuery type."""
     if property_format == "date-time":
         return "timestamp"
