@@ -96,7 +96,6 @@ class GcsStagingProcessWorker(GcsStagingWorker, Process):
 
 
 class BigQueryGcsStagingSink(BaseBigQuerySink):
-
     MAX_WORKERS = os.cpu_count() * 2
     WORKER_CAPACITY_FACTOR = 1
     WORKER_CREATION_MIN_INTERVAL = 10.0
@@ -167,6 +166,7 @@ class BigQueryGcsStagingSink(BaseBigQuerySink):
             self.logger.info("Data loaded successfully")
         else:
             self.logger.info("No data to load")
+        super().clean_up()
 
 
 class BigQueryGcsStagingDenormalizedSink(Denormalized, BigQueryGcsStagingSink):
@@ -179,6 +179,7 @@ class BigQueryGcsStagingDenormalizedSink(Denormalized, BigQueryGcsStagingSink):
             "schema_update_options": [
                 bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION,
             ],
+            "ignore_unknown_values": True,
         }
 
     # Defer schema evolution the the write disposition
